/**
 * Created by Tobi Stockinger on 26/11/14.
 */
var express = require('express');
var router = express.Router();
var debug = require('debug')('assignment07');
// status codes
var codes = require('../data/codes.js');


String.prototype.hashCode = function() {
    var hash = 0, i, chr, len;
    if (this.length == 0) return hash;
    for (i = 0, len = this.length; i < len; i++) {
        chr   = this.charCodeAt(i);
        hash  = ((hash << 5) - hash) + chr;
        hash |= 0; // Convert to 32bit integer
    }
    return hash;
};

/**
 * this constructs a new suggestion that we can insert into the database
 * @param id
 * @param text
 * @constructor
 */
var Suggestion = function(text){
    this.id = -1;
    this.text = text;
    this.votes = {
        up: 0,
        down: 0
    };
    this.votesByUser = {};
};

/**
 * tries to determine which ip address
 * the request originates from
 * @param request
 * @returns {*|Socket.remoteAddress|CleartextStream.remoteAddress}
 */
var getIpForRequest = function(request){
    return request.headers['x-forwarded-for'] ||
        request.connection.remoteAddress ||
        request.socket.remoteAddress ||
        request.connection.socket.remoteAddress;
};

/**
 * mongo does not allow us to insert keys that contain dots
 * since we want to identify users according to their IP address,
 * we need to either remove the dots (that works), or
 * hash the ip (which is better in terms of privacy)
 * @param ip
 * @returns {*}
 */
var getUserHash = function(ip){
    var ipString = ip.toString();
    return ip.hashCode();
};


/* GET suggestions as list */
router.get('/suggestions',function(request,response){
    /**
     * the score is simple the number of positive votes minus the negative votes
     * @param suggestion
     * @returns {number}
     */
    var getScore = function(suggestion){
        return suggestion.votes.up - suggestion.votes.down;
    };
    var db = request.db;
    var suggestions = db.get('suggestions');
    var ip = getUserHash(getIpForRequest(request));
    suggestions.find({},{},function(e,docs){

        var responseSuggestions = [], suggestion, loopSuggestion;

        // sort the array with a custom sort function
        docs.sort(function(a,b){
            // the score is simply upVotes - downVotes;
            return getScore(b) - getScore(a);
        });

        // loop through all global suggestions
        // and construct a snapshot of the values.
        // important: omit the 'votesByUser' key,
        // because we don't want to reveal who has voted.
        for(var i=0;i<docs.length;i++){
            suggestion = docs[i];

            loopSuggestion = {};
            loopSuggestion.hasVoted = false;
            // let's see if the user has already voted
            // and set the according flag in the suggestion
            if(typeof suggestion.votesByUser[ip] != 'undefined'){
                loopSuggestion.hasVoted = true;
                loopSuggestion.ownVote = suggestion.votesByUser[ip];
            }


            // copy the values to our public response object
            loopSuggestion.id = suggestion.id;
            loopSuggestion.text = suggestion.text;
            loopSuggestion.votes = suggestion.votes;

            responseSuggestions.push(loopSuggestion);
        }

        response.json({
            status : codes.SUCCESS.code,
            message : codes.SUCCESS.message,
            suggestions : responseSuggestions
        })

    });
});

/* POST to home page. */
router.post('/', function(request, response) {
    var method, clientIP, affectedSuggestion;
    var responseObj = {
        status : -1,
        message : '',
        suggestion : {}
    };

    var db = request.db;
    var suggestions = db.get('suggestions');

    var queryObj;
    var parsedId, objId;

    var successCallback, errorCallback;

    if(request.body){
        console.log(request.body);

        parsedId = parseInt(request.body.suggestionId);
        objId =  suggestions.id('5488102ed1427ba81e18954c');
        console.log("Object id:");
        console.log(objId);

        queryObj = {$or : [{id : parsedId},{_id : objId }]};

        // currently 'up' or 'down', but potentially everything can be passed and will be stored here.
        method = request.body.method;

        // from which IP came the request?
        clientIP = getUserHash(getIpForRequest(request));

        successCallback = function(error,docs){
            affectedSuggestion = docs;

            try{
                // first, let's check if there already an key value pair
                // for this IP for the suggestion.
                // if there is, we need to prevent the user from voting.
                // Also, return an error/message object
                if(affectedSuggestion.votesByUser[clientIP]){

                    // the user tries to vote again for the same suggestion
                    // the UI should prevent that, but if not, let's fail gracefully.
                    if(method == affectedSuggestion.votesByUser[clientIP]){
                        response.json(codes.ALREADY_VOTED);
                        return;
                    }

                    // user has changed his mind. we need to decrement the votes and delete the field;
                    else{
                        // let's see how the user had voted initially
                        // then decrement the overall votes for that.
                        affectedSuggestion.votes[affectedSuggestion.votesByUser[clientIP]]--;

                        // delete the user's vote
                        delete affectedSuggestion.votesByUser[clientIP];
                    }
                }

                // put the clientIP into the list of votesByUser
                // that allows us to prevent double-votes.
                affectedSuggestion.votesByUser[clientIP] = method;

                // increment the counter for that suggestion
                affectedSuggestion.votes[method]++;

                // now we've altered the affectedSuggestion and need to store it in the database
                suggestions.updateById(affectedSuggestion._id.toString(),affectedSuggestion,function(error,data){
                    if(data){
                        // once we get here, everything worked as expected
                        // and we can construct the response object
                        responseObj = {
                            status : codes.COUNT_UPDATED.code,
                            message : codes.COUNT_UPDATED.message,

                            // add the most important properties of the affected Suggestion to the response
                            suggestion : {
                                id : affectedSuggestion.id,
                                votes : affectedSuggestion.votes,
                                ownVote : method
                            }
                        };
                    }else if(error){
                        // once we get here, everything worked as expected
                        // and we can construct the response object
                        responseObj = codes.DB_UPDATE_ERROR;
                    }

                    // TODO emit a socket.io event with the same data.
                    response.json(responseObj);
                });
            }catch (e){
                response.json(codes.NO_SUCH_SUGGESTION);
            }
        };
        errorCallback = function(error,docs){
          response.json(codes.NO_SUCH_SUGGESTION);
        };

        // locate the suggestion by its id. make sure to parse it first, because the body parser makes a string out of it.
        suggestions.findOne(queryObj,{},
            function(error,docs){
                if(docs){
                    successCallback(error,docs);
                }
                else{
                    suggestions.findById(request.body.suggestionId,function(error,docs){
                        if(error || !docs){
                            errorCallback(error,docs);
                        }
                        else{
                            successCallback(error,docs);
                        }
                    });
                }
            }
        );
    }
    else{
        responseObj = codes.EMPTY_REQUEST;
        // this is where we send the actual response.
        response.json(responseObj);
    }

});


/* POST to insert a new suggestion */
router.post('/suggestions',function(request,response){
    var db = request.db;
    var suggestions = db.get('suggestions');

    var ip = getUserHash(getIpForRequest(request));


    console.log(request.body);
    if(request.body && request.body.suggestionText){
        var suggestion = new Suggestion(request.body.suggestionText);

        // the user that suggests automatically votes 'up'
        suggestion.votes.up++;
        suggestion.votesByUser[ip] = 'up';

        suggestions.insert(suggestion,function(error,data){
            if(data){
                // in the front-end we refer to the suggestion via the 'id' property, not '_id'
                // that is why we update this property and THEN send the reponse.
                data.id = data._id.toString();
                data.ownVote = 'up';
                data.hasVoted = true;
                suggestions.updateById(data.id,data,function(){
                    delete data.votesByUser; // do not send this to the client
                    delete data._id;
                    response.json({
                        status : codes.SUCCESS.code,
                        message : codes.SUCCESS.message,
                        suggestions : data
                    });
                });
            }
            else{
                response.json(codes.DB_UPDATE_ERROR);
            }
        });
    }
    else{
        response.json(codes.EMPTY_REQUEST);
    }
});


/**
 * this is a handy endpoint to reset all suggestions.
 * POST Parameters: {really: 'really'}
 */
router.post('/clearSuggestions',
    function(request,response){
        var db = request.db;
        var suggestions = db.get('suggestions');
        if(request.body && request.body.really == 'really'){
            suggestions.remove({},function(err,data){
                response.json(codes.DB_FLUSHED_SUCCESS);
                suggestions.insert(require('../data/data.js'),{},function(error,docs){
                    if(docs){
                        console.log('successfully reset data');
                    }
                    else{
                        console.log('could not reset data in database');
                    }
                });
            });
        }
        else{
            response.json(codes.ERROR_WRONG_PARAMETERS);
        }
    });

module.exports = router;